"""Emoji used by MILC when outputting logs

| Log Level | Emoji |
|-----------|-------|
| `CRITICAL` | `{bg_red}{fg_white}¬_¬` |
| `ERROR` | `{fg_red}☒` |
| `WARNING` | `{fg_yellow}⚠` |
| `INFO` | `{fg_blue}ℹ` |
| `DEBUG` | `{fg_cyan}☐` |
| `NOTSET` | `{style_reset_all}¯\\_(o_o)_/¯` |

If you'd like to use your own icon for a level instead you can simply redefine it:

```python
from milc.emoji import EMOJI_LOGLEVELS

EMOJI_LOGLEVELS['INFO'] = {fg_green}ℹ'
```
"""
EMOJI_LOGLEVELS = {
    'CRITICAL': '{bg_red}{fg_white}¬_¬',
    'ERROR': '{fg_red}☒',
    'WARNING': '{fg_yellow}⚠',
    'INFO': '{fg_blue}ℹ',
    'DEBUG': '{fg_cyan}☐',
    'NOTSET': '{style_reset_all}¯\\_(o_o)_/¯',
}
EMOJI_LOGLEVELS['FATAL'] = EMOJI_LOGLEVELS['CRITICAL']
EMOJI_LOGLEVELS['WARN'] = EMOJI_LOGLEVELS['WARNING']



"""Sometimes you need to ask the user a question. MILC provides basic functions for collecting and validating user input. You can find these in the `milc.questions` module.
"""
from getpass import getpass

import milc
from .ansi import format_ansi


def yesno(prompt, *args, default=None, **kwargs):
    """Displays `prompt` to the user and gets a yes or no response.

    Returns `True` for a yes and `False` for a no.

    | Argument | Description |
    |----------|-------------|
    | prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | default | Whether to default to a Yes or No when the user presses enter.<br><br>None- force the user to enter Y or N<br>True- Default to yes<br>False- Default to no |

    If you add `--yes` and `--no` arguments to your program the user can answer questions by passing command line flags.

    ```python
    @cli.argument('-y', '--yes', action='store_true', arg_only=True, help='Answer yes to all questions.')
    @cli.argument('-n', '--no', action='store_true', arg_only=True, help='Answer no to all questions.')
    ```
    """
    if not args and kwargs:
        args = kwargs

    if 'no' in milc.cli.args and milc.cli.args.no:
        return False

    if 'yes' in milc.cli.args and milc.cli.args.yes:
        return True

    if not milc.cli.interactive:
        return False

    if default is None:
        prompt = prompt + ' [y/n] '
    elif default:
        prompt = prompt + ' [Y/n] '
    else:
        prompt = prompt + ' [y/N] '

    while True:
        answer = input(format_ansi(prompt % args))

        if not answer and default is not None:
            return default

        elif answer.lower() in ['y', 'yes']:
            return True

        elif answer.lower() in ['n', 'no']:
            return False


def password(prompt='Enter password:', *args, confirm=False, confirm_prompt='Confirm password:', confirm_limit=3, validate=None, **kwargs):
    """Securely receive a password from the user. Returns the password or None.

    | Argument | Description |
    |----------|-------------|
    | prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | confirm | Prompt the user to type the password again and make sure they match. |
    | confirm_prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | confirm_limit | Number of attempts to confirm before giving up. Default: 3 |
    | validate | This is an optional function that can be used to validate the password, EG to check complexity. It should return True or False and have the following signature:<br><br>`def function_name(answer):` |
    """
    if not milc.cli.interactive:
        return None

    if not args and kwargs:
        args = kwargs

    if prompt[-1] != ' ':
        prompt += ' '

    if confirm_prompt[-1] != ' ':
        confirm_prompt += ' '

    i = 0
    while not confirm_limit or i < confirm_limit:
        pw = getpass(format_ansi(prompt % args))

        if pw:
            if validate is not None and not validate(pw):
                continue

            elif confirm:
                if getpass(format_ansi(confirm_prompt % args)) == pw:
                    return pw
                else:
                    milc.cli.log.error('Passwords do not match!')

            else:
                return pw

            i += 1


def question(prompt, *args, default=None, confirm=False, answer_type=str, validate=None, **kwargs):
    """Allow the user to type in a free-form string to answer.

    | Argument | Description |
    |----------|-------------|
    | prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | default | The value to return when the user doesn't enter any value. Use None to prompt until they enter a value. |
    | confirm | Present the user with a confirmation dialog before accepting their answer. |
    | answer_type | Specify a type function for the answer. Will re-prompt the user if the function raises any errors. Common choices here include int, float, and decimal.Decimal. |
    | validate | This is an optional function that can be used to validate the answer. It should return True or False and have the following signature:<br><br>`def function_name(answer, *args, **kwargs):` |
    """
    if not milc.cli.interactive:
        return default

    if default is not None:
        prompt = '%s [%s] ' % (prompt, default)
    elif prompt and prompt[-1] != ' ':
        prompt += ' '

    while True:
        answer = input(format_ansi(prompt % (args or kwargs)))

        if answer:
            if validate is not None and not validate(answer, *args, **kwargs):
                continue

            elif confirm:
                if yesno('Is the answer "%s" correct?', answer, default=True):
                    try:
                        return answer_type(answer)
                    except Exception as e:
                        milc.cli.log.error('Could not convert answer (%s) to type %s: %s', answer, answer_type.__name__, str(e))

            else:
                try:
                    return answer_type(answer)
                except Exception as e:
                    milc.cli.log.error('Could not convert answer (%s) to type %s: %s', answer, answer_type.__name__, str(e))

        elif default is not None:
            return default


def choice(heading, options, *args, default=None, confirm=False, prompt='Please enter your choice: ', **kwargs):
    """Present the user with a list of options and let them select one.

    Users can enter either the number or the text of their choice. This will return the value of the item they choose, not the numerical index.

    | Argument | Description |
    |----------|-------------|
    | heading | The text to place above the list of options. |
    | options | A sequence of items to choose from. |
    | default | The index of the item to return when the user doesn't enter any value. Use None to prompt until they enter a value. |
    | confirm | When True present the user with a confirmation dialog before accepting their answer. |
    | prompt | The prompt to present to the user. Can include color and format strings like milc's `cli.echo()`. |

    Users can enter either the number or the text of their choice.

    !!! warning
        This will return the value of the item they choose, not the numerical index.
    """
    if not args and kwargs:
        args = kwargs

    if not milc.cli.interactive:
        return default

    if prompt and default is not None:
        prompt = prompt + ' [%s] ' % (default + 1,)
    elif prompt[-1] != ' ':
        prompt += ' '

    while True:
        # Prompt for an answer.
        milc.cli.echo(heading % args)
        for i, option in enumerate(options, 1):
            milc.cli.echo('\t{fg_cyan}%d.{fg_reset} %s', i, option)

        answer = input(format_ansi(prompt))

        # If the user types in one of the options exactly use that
        if answer in options:
            return answer

        # Massage the answer into a valid integer
        if answer == '' and default is not None:
            answer = default
        else:
            try:
                answer = int(answer) - 1
            except Exception as e:
                milc.cli.log.error('Invalid choice: %s', answer)
                milc.cli.log.debug('Could not convert %s to int: %s: %s', answer, e.__class__.__name__, e)
                if milc.cli.config.general.verbose:
                    milc.cli.log.exception(e)
                continue

        # Validate the answer
        if answer >= len(options) or answer < 0:
            milc.cli.log.error('Invalid choice: %s', answer + 1)
            continue

        if confirm and not yesno('Is the answer "%s" correct?', answer + 1, default=True):
            continue

        # Return the answer they chose.
        return options[answer]



"""Read and write configuration settings
"""
import milc


def print_config(section, key):
    """Print a single config setting to stdout.
    """
    if milc.cli.config_source[section][key] == 'config_file':
        milc.cli.echo('%s.%s{fg_blue}={fg_reset}%s', section, key, milc.cli.config[section][key])
    else:
        milc.cli.echo('{fg_cyan}%s.%s=%s', section, key, milc.cli.config[section][key])


def show_config():
    """Print the current configuration to stdout.
    """
    for section in sorted(milc.cli.config):
        for key in sorted(milc.cli.config[section]):
            if milc.cli.config_source[section][key] == 'config_file' or milc.cli.config.config.all:
                print_config(section, key)


def parse_config_token(config_token):
    """Split a user-supplied configuration-token into its components.
    """
    section = option = value = None

    if '=' in config_token and '.' not in config_token:
        milc.cli.log.error('Invalid configuration token, the key must be of the form <section>.<option>: %s', config_token)
        return section, option, value

    # Separate the key (<section>.<option>) from the value
    if '=' in config_token:
        key, value = config_token.split('=')
    else:
        key = config_token

    # Extract the section and option from the key
    if '.' in key:
        section, option = key.split('.', 1)
    else:
        section = key

    return section, option, value


def set_config(section, option, value):
    """Set a config key in the running config.
    """
    log_string = '%s.%s{fg_cyan}:{fg_reset} %s {fg_cyan}->{fg_reset} %s'
    if milc.cli.args.read_only:
        log_string += ' {fg_red}(change not written)'

    milc.cli.echo(log_string, section, option, milc.cli.config[section][option], value)

    if not milc.cli.args.read_only:
        if value == 'None':
            del milc.cli.config[section][option]
        else:
            milc.cli.config[section][option] = value
            milc.cli.config_source[section][option] = 'config_file'


@milc.cli.argument('-a', '--all', action='store_true', help='Show all configuration options.')
@milc.cli.argument('-ro', '--read-only', arg_only=True, action='store_true', help='Operate in read-only mode.')
@milc.cli.argument('configs', nargs='*', arg_only=True, help='Configuration options to read or write.')
@milc.cli.subcommand("Read and write configuration settings.")
def config(cli):
    """Read and write config settings.

    This script iterates over the config_tokens supplied as argument. Each config_token has the following form:

        section[.key][=value]

    If only a section (EG 'compile') is supplied all keys for that section will be displayed.

    If section.key is supplied the value for that single key will be displayed.

    If section.key=value is supplied the value for that single key will be set.

    If section.key=None is supplied the key will be deleted.

    No validation is done to ensure that the supplied section.key is actually used by a subcommand.
    """
    if not milc.cli.args.configs:
        return show_config()

    # Process config_tokens
    save_config = False

    for config_token in milc.cli.args.configs:
        section, option, value = parse_config_token(config_token)

        # Validation
        if option and '.' in option:
            milc.cli.log.error('Config keys may not have more than one period! "%s" is not valid.', config_token)
            return False

        # Do what the user wants
        if section and option and value:
            # Write a configuration option
            set_config(section, option, value)
            if not milc.cli.args.read_only:
                save_config = True

        elif section and option:
            # Display a single key
            print_config(section, option)

        elif section:
            # Display an entire section
            for key in milc.cli.config[section]:
                if milc.cli.config_source[section][key] == 'config_file':
                    print_config(section, key)

    # Ending actions
    if save_config:
        milc.cli.save_config()

    return True



#!/usr/bin/env python3
# coding=utf-8
"""MILC - A CLI Framework

PYTHON_ARGCOMPLETE_OK

MILC is an opinionated framework for writing CLI apps. It optimizes for the
most common unix tool pattern- small tools that are run from the command
line and generally do not feature any user interaction while they run.

For more details see the MILC documentation:

    <https://github.com/clueboard/milc/tree/master/docs>
"""
__VERSION__ = '1.6.2'

import logging
import os
import sys
import warnings

from .emoji import EMOJI_LOGLEVELS
from .milc import MILC


def argv_name():
    """Returns the name of our program by examining argv.
    """
    app_name = sys.argv[0][:-3] if sys.argv[0].endswith('.py') else sys.argv[0]
    return os.path.split(app_name)[-1]


APP_NAME = os.environ.get('MILC_APP_NAME') or argv_name()
APP_VERSION = os.environ.get('MILC_APP_VERSION', 'unknown')
APP_AUTHOR = os.environ.get('MILC_APP_AUTHOR', APP_NAME.upper())

if 'MILC_IGNORE_DEPRECATED' not in os.environ:
    for name in ('MILC_APP_NAME', 'MILC_APP_VERSION', 'MILC_APP_AUTHOR'):
        if name in os.environ:
            warnings.warn(f'Using {name} is deprecated and will not be supported in the future, please use set_metadata() instead.', stacklevel=2)

logging.basicConfig(stream=os.devnull)  # Disable logging until we can configure it how the user wants

cli = MILC(APP_NAME, APP_VERSION, APP_AUTHOR)


def set_metadata(*, name=APP_NAME, author=APP_AUTHOR, version=APP_VERSION):
    """Set metadata about your program.

    This allows you to set the application's name, version, and/or author
    before executing your entrypoint. It's best to run this only once, and
    it must be run before you call `cli()`.
    """
    global APP_NAME, APP_VERSION, APP_AUTHOR, cli

    if cli._inside_context_manager:
        raise RuntimeError('You must run set_metadata() before cli()!')

    APP_NAME = name
    APP_VERSION = version
    APP_AUTHOR = author
    cli = MILC(name, version, author)


# Extra stuff people can import
from ._sparkline import sparkline  # noqa



class AttrDict(object):
    """A dictionary that can also be accessed by attribute.
    """
    def __contains__(self, key):
        return self._data.__contains__(key)

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()

    def __repr__(self):
        return self._data.__repr__()

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    def values(self):
        return self._data.values()

    def __init__(self, *args, **kwargs):
        self._data = {}

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __getitem__(self, key):
        """Returns an item.
        """
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        self.__setattr__(key, value)

    def __delitem__(self, key):
        if key in self._data:
            del self._data[key]


class SparseAttrDict(AttrDict):
    """A dictionary that can also be accessed by attribute.

    This class never raises IndexError, instead it will return None if a
    key does not yet exist.
    """
    def __getitem__(self, key):
        """Returns an item, creating it if it doesn't already exist
        """
        if key not in self._data:
            self._data[key] = None

        return self._data[key]
